import typing

try:
    from inspect import getfullargspec as _getargspec
except ImportError:
    from inspect import getargspec as _getargspec

import copy as _copy
import hashlib as _hashlib
import json as _json
import os as _os
import sys as _sys

import six as _six
from google.protobuf.json_format import MessageToDict as _MessageToDict

from flytekit.bin import entrypoint as _entrypoint
from flytekit.common import constants as _constants
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.tasks import output as _task_output
from flytekit.common.tasks import sdk_runnable as _sdk_runnable
from flytekit.common.types import helpers as _type_helpers
from flytekit.models import literals as _literal_models
from flytekit.models import task as _task_models
from flytekit.plugins import pyspark as _pyspark


class GlobalSparkContext(object):
    _SPARK_CONTEXT = None
    _SPARK_SESSION = None

    @classmethod
    def get_spark_context(cls):
        return cls._SPARK_CONTEXT

    @classmethod
    def get_spark_session(cls):
        return cls._SPARK_SESSION

    def __enter__(self):
        GlobalSparkContext._SPARK_CONTEXT = _pyspark.SparkContext()
        GlobalSparkContext._SPARK_SESSION = _pyspark.sql.SparkSession.builder.appName(
            "Flyte Spark SQL Context"
        ).getOrCreate()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        GlobalSparkContext._SPARK_CONTEXT.stop()
        GlobalSparkContext._SPARK_CONTEXT = None
        return False


class SdkRunnableSparkContainer(_sdk_runnable.SdkRunnableContainer):
    @property
    def args(self):
        """
        Override args to remove the injection of command prefixes
        :rtype: list[Text]
        """
        return self._args


class SdkSparkTask(_sdk_runnable.SdkRunnableTask):
    """
    This class includes the additional logic for building a task that executes as a Spark Job.

    """

    def __init__(
        self,
        task_function,
        task_type,
        discovery_version,
        retries,
        interruptible,
        deprecated,
        discoverable,
        timeout,
        spark_type,
        spark_conf,
        hadoop_conf,
        environment,
    ):
        """
        :param task_function: Function container user code.  This will be executed via the SDK's engine.
        :param Text task_type: string describing the task type
        :param Text discovery_version: string describing the version for task discovery purposes
        :param int retries: Number of retries to attempt
        :param bool interruptible: Whether or not task is interruptible
        :param Text deprecated:
        :param bool discoverable:
        :param datetime.timedelta timeout:
        :param dict[Text,Text] spark_conf:
        :param dict[Text,Text] hadoop_conf:
        :param dict[Text,Text] environment: [optional] environment variables to set when executing this task.
        """

        spark_exec_path = _os.path.abspath(_entrypoint.__file__)
        if spark_exec_path.endswith(".pyc"):
            spark_exec_path = spark_exec_path[:-1]

        self._spark_job = _task_models.SparkJob(
            spark_conf=spark_conf,
            hadoop_conf=hadoop_conf,
            application_file="local://" + spark_exec_path,
            executor_path=_sys.executable,
            main_class="",
            spark_type=spark_type,
        )
        super(SdkSparkTask, self).__init__(
            task_function,
            task_type,
            discovery_version,
            retries,
            interruptible,
            deprecated,
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            "",
            discoverable,
            timeout,
            environment,
            _MessageToDict(self._spark_job.to_flyte_idl()),
        )

    @_exception_scopes.system_entry_point
    def execute(self, context, inputs):
        """
        :param flytekit.engines.common.EngineContext context:
        :param flytekit.models.literals.LiteralMap inputs:
        :rtype: dict[Text,flytekit.models.common.FlyteIdlEntity]
        :returns: This function must return a dictionary mapping 'filenames' to Flyte Interface Entities.  These
            entities will be used by the engine to pass data from node to node, populate metadata, etc. etc..  Each
            engine will have different behavior.  For instance, the Flyte engine will upload the entities to a remote
            working directory (with the names provided), which will in turn allow Flyte Propeller to push along the
            workflow.  Where as local engine will merely feed the outputs directly into the next node.
        """
        inputs_dict = _type_helpers.unpack_literal_map_to_sdk_python_std(
            inputs,
            {k: _type_helpers.get_sdk_type_from_literal_type(v.type) for k, v in _six.iteritems(self.interface.inputs)},
        )
        outputs_dict = {
            name: _task_output.OutputReference(_type_helpers.get_sdk_type_from_literal_type(variable.type))
            for name, variable in _six.iteritems(self.interface.outputs)
        }

        inputs_dict.update(outputs_dict)

        with GlobalSparkContext():
            _exception_scopes.user_entry_point(self.task_function)(
                _sdk_runnable.ExecutionParameters(
                    execution_date=context.execution_date,
                    tmp_dir=context.working_directory,
                    stats=context.stats,
                    execution_id=context.execution_id,
                    logging=context.logging,
                ),
                GlobalSparkContext.get_spark_context(),
                **inputs_dict
            )
        return {
            _constants.OUTPUT_FILE_NAME: _literal_models.LiteralMap(
                literals={k: v.sdk_value for k, v in _six.iteritems(outputs_dict)}
            )
        }

    @property
    def spark_conf(self):
        return self._spark_job.spark_conf

    @property
    def hadoop_conf(self):
        return self._spark_job.hadoop_conf

    def _get_container_definition(self, **kwargs):
        """
        :rtype: SdkRunnableSparkContainer
        """
        return super(SdkSparkTask, self)._get_container_definition(cls=SdkRunnableSparkContainer, **kwargs)

    def _get_kwarg_inputs(self):
        # Trim off first two parameters as they are reserved for workflow_parameters and spark_context
        return set(_getargspec(self.task_function).args[2:])

    def with_overrides(
        self, new_spark_conf: typing.Dict[str, str] = None, new_hadoop_conf: typing.Dict[str, str] = None
    ):
        """
        Creates a new SparkJob instance with the modified configuration or timeouts
        """
        tk = _copy.deepcopy(self)
        tk._spark_job = self._spark_job.with_overrides(new_spark_conf, new_hadoop_conf)
        tk._custom = _MessageToDict(tk._spark_job.to_flyte_idl())

        salt = _hashlib.md5(_json.dumps(tk.custom, sort_keys=True).encode("utf-8")).hexdigest()
        tk._id._name = "{}-{}".format(self._id.name, salt)
        # We are overriding the platform name creation to prevent problems in dynamic
        tk.assign_name(tk._id._name)

        return tk



import six as _six

from flytekit.common import promise as _promise
from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import containers as _containers
from flytekit.common.types import helpers as _type_helpers
from flytekit.common.types import primitives as _primitives
from flytekit.models import interface as _interface_models
from flytekit.models import literals as _literal_models


class BindingData(_literal_models.BindingData, metaclass=_sdk_bases.ExtendedSdkType):
    @staticmethod
    def _has_sub_bindings(m):
        """
        :param dict[Text,T] or list[T]:
        :rtype: bool
        """
        for v in _six.itervalues(m) if isinstance(m, dict) else m:
            if isinstance(v, (list, dict)) and BindingData._has_sub_bindings(v):
                return True
            elif isinstance(v, (_promise.Input, _promise.NodeOutput)):
                return True
        return False

    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.literals.BindingData model:
        :rtype: BindingData
        """
        return cls(
            scalar=model.scalar,
            collection=model.collection,
            promise=model.promise,
            map=model.map,
        )

    @classmethod
    def from_python_std(cls, literal_type, t_value, upstream_nodes=None):
        """
        :param flytekit.models.types.LiteralType literal_type:
        :param T t_value:
        :param list[flytekit.common.nodes.SdkNode] upstream_nodes: [Optional] Keeps track of the nodes upstream,
            if applicable.
        :rtype: BindingData
        """
        scalar = None
        collection = None
        promise = None
        map = None
        downstream_sdk_type = _type_helpers.get_sdk_type_from_literal_type(literal_type)
        if isinstance(t_value, _promise.Input):
            if not downstream_sdk_type.is_castable_from(t_value.sdk_type):
                _user_exceptions.FlyteTypeException(
                    t_value.sdk_type,
                    downstream_sdk_type,
                    additional_msg="When binding workflow input: {}".format(t_value),
                )
            promise = t_value.promise
        elif isinstance(t_value, _promise.NodeOutput):
            if not downstream_sdk_type.is_castable_from(t_value.sdk_type):
                _user_exceptions.FlyteTypeException(
                    t_value.sdk_type,
                    downstream_sdk_type,
                    additional_msg="When binding node output: {}".format(t_value),
                )
            promise = t_value
            if upstream_nodes is not None:
                upstream_nodes.append(t_value.sdk_node)
        elif isinstance(t_value, list):
            if not issubclass(downstream_sdk_type, _containers.ListImpl):
                raise _user_exceptions.FlyteTypeException(
                    type(t_value),
                    downstream_sdk_type,
                    received_value=t_value,
                    additional_msg="Cannot bind a list to a non-list type.",
                )
            collection = _literal_models.BindingDataCollection(
                [
                    BindingData.from_python_std(
                        downstream_sdk_type.sub_type.to_flyte_literal_type(),
                        v,
                        upstream_nodes=upstream_nodes,
                    )
                    for v in t_value
                ]
            )
        elif isinstance(t_value, dict) and (
            not issubclass(downstream_sdk_type, _primitives.Generic) or BindingData._has_sub_bindings(t_value)
        ):
            # TODO: This behavior should be embedded in the type engine.  Someone should be able to alter behavior of
            # TODO: binding logic by injecting their own type engine.  The same goes for the list check above.
            raise NotImplementedError("TODO: Cannot use map bindings at the moment")
        else:
            sdk_value = downstream_sdk_type.from_python_std(t_value)
            scalar = sdk_value.scalar
            collection = sdk_value.collection
            map = sdk_value.map
        return cls(scalar=scalar, collection=collection, map=map, promise=promise)


class TypedInterface(_interface_models.TypedInterface, metaclass=_sdk_bases.ExtendedSdkType):
    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.interface.TypedInterface model:
        :rtype: TypedInterface
        """
        return cls(model.inputs, model.outputs)

    def create_bindings_for_inputs(self, map_of_bindings):
        """
        :param dict[Text, T] map_of_bindings:  This can be scalar primitives, it can be node output references,
            lists, etc..
        :rtype: (list[flytekit.models.literals.Binding], list[flytekit.common.nodes.SdkNode])
        :raises: flytekit.common.exceptions.user.FlyteAssertion
        """
        binding_data = dict()
        all_upstream_nodes = list()
        for k in sorted(self.inputs):
            var = self.inputs[k]
            if k not in map_of_bindings:
                raise _user_exceptions.FlyteAssertion("Input was not specified for: {} of type {}".format(k, var.type))

            binding_data[k] = BindingData.from_python_std(
                var.type, map_of_bindings[k], upstream_nodes=all_upstream_nodes
            )

        extra_inputs = set(binding_data.keys()) ^ set(map_of_bindings.keys())
        if len(extra_inputs) > 0:
            raise _user_exceptions.FlyteAssertion(
                "Too many inputs were specified for the interface.  Extra inputs were: {}".format(extra_inputs)
            )

        seen_nodes = set()
        min_upstream = list()
        for n in all_upstream_nodes:
            if n not in seen_nodes:
                seen_nodes.add(n)
                min_upstream.append(n)

        return (
            [_literal_models.Binding(k, bd) for k, bd in _six.iteritems(binding_data)],
            min_upstream,
        )

    def __repr__(self):
        return "({inputs}) -> ({outputs})".format(
            inputs=", ".join(
                [
                    "{}: {}".format(k, _type_helpers.get_sdk_type_from_literal_type(v.type))
                    for k, v in _six.iteritems(self.inputs)
                ]
            ),
            outputs=", ".join(
                [
                    "{}: {}".format(k, _type_helpers.get_sdk_type_from_literal_type(v.type))
                    for k, v in _six.iteritems(self.outputs)
                ]
            ),
        )



import datetime as _datetime
import json as _json

import six as _six
from dateutil import parser as _parser
from google.protobuf import json_format as _json_format
from google.protobuf import struct_pb2 as _struct
from pytimeparse import parse as _parse_duration_string

from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.models import literals as _literals
from flytekit.models import types as _idl_types


class Integer(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Integer
        """
        try:
            return cls(int(string_value))
        except (ValueError, TypeError):
            raise _user_exceptions.FlyteTypeException(
                _six.text_type,
                int,
                additional_msg="String not castable to Integer SDK type:" " {}".format(string_value),
            )

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        if type(t_value) not in _six.integer_types:
            raise _user_exceptions.FlyteTypeException(type(t_value), _six.integer_types, t_value)
        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.INTEGER)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Integer
        """
        return cls(literal_model.scalar.primitive.integer)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Integer"

    def __init__(self, value):
        """
        :param int value: Int value to wrap
        """
        super(Integer, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(integer=value)))

    def to_python_std(self):
        """
        :rtype: int
        """
        return self.scalar.primitive.integer

    def short_string(self):
        """
        :rtype: Text
        """
        return "Integer({})".format(self.scalar.primitive.integer)


class Float(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Float
        """
        try:
            return cls(float(string_value))
        except ValueError:
            raise _user_exceptions.FlyteTypeException(
                _six.text_type,
                float,
                additional_msg="String not castable to Float SDK type:" " {}".format(string_value),
            )

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        if type(t_value) != float:
            raise _user_exceptions.FlyteTypeException(type(t_value), float, t_value)
        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.FLOAT)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Float
        """
        return cls(literal_model.scalar.primitive.float_value)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Float"

    def __init__(self, value):
        """
        :param float value: value to wrap
        """
        super(Float, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(float_value=value)))

    def to_python_std(self):
        """
        :rtype: float
        """
        return self.scalar.primitive.float_value

    def short_string(self):
        """
        :rtype: Text
        """
        return "Float({})".format(self.scalar.primitive.float_value)


class Boolean(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Boolean
        """
        if string_value == "1" or string_value.lower() == "true":
            return cls(True)
        elif string_value == "0" or string_value.lower() == "false":
            return cls(False)
        raise _user_exceptions.FlyteTypeException(
            _six.text_type,
            bool,
            additional_msg="String not castable to Boolean SDK " "type: {}".format(string_value),
        )

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        if type(t_value) != bool:
            raise _user_exceptions.FlyteTypeException(type(t_value), bool, t_value)
        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.BOOLEAN)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: bool
        """
        return cls(literal_model.scalar.primitive.boolean)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Boolean"

    def __init__(self, value):
        """
        :param bool value: value to wrap
        """
        super(Boolean, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(boolean=value)))

    def to_python_std(self):
        """
        :rtype: bool
        """
        return self.scalar.primitive.boolean

    def short_string(self):
        """
        :rtype: Text
        """
        return "Boolean({})".format(self.scalar.primitive.boolean)


class String(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: String
        """
        if type(string_value) == dict or type(string_value) == list:
            raise _user_exceptions.FlyteTypeException(
                type(string_value),
                _six.text_type,
                additional_msg="Should not cast native Python type to string {}".format(string_value),
            )
        return cls(string_value)

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        Creates an object of this type from the model primitive defining it.
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        if type(t_value) not in set([str, _six.text_type]):
            raise _user_exceptions.FlyteTypeException(type(t_value), set([str, _six.text_type]), t_value)
        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.STRING)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: String
        """
        return cls(literal_model.scalar.primitive.string_value)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "String"

    def __init__(self, value):
        """
        :param Text value: value to wrap
        """
        super(String, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(string_value=value)))

    def to_python_std(self):
        """
        :rtype: Text
        """
        return self.scalar.primitive.string_value

    def short_string(self):
        """
        :rtype: Text
        """
        _TRUNCATE_LENGTH = 100
        return "String('{}'{})".format(
            self.scalar.primitive.string_value[:_TRUNCATE_LENGTH],
            " ..." if len(self.scalar.primitive.string_value) > _TRUNCATE_LENGTH else "",
        )

    def verbose_string(self):
        """
        :rtype: Text
        """
        return "String('{}')".format(self.scalar.primitive.string_value)


class Datetime(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value:
        :rtype: Datetime
        """
        try:
            python_std_datetime = _parser.parse(string_value)
        except ValueError:
            raise _user_exceptions.FlyteTypeException(
                _six.text_type,
                _datetime.datetime,
                additional_msg="String not castable to Datetime " "SDK type: {}".format(string_value),
            )

        return cls.from_python_std(python_std_datetime)

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif type(t_value) != _datetime.datetime:
            raise _user_exceptions.FlyteTypeException(type(t_value), _datetime.datetime, t_value)
        elif t_value.tzinfo is None:
            raise _user_exceptions.FlyteValueException(
                t_value,
                "Datetime objects in Flyte must be timezone aware.  " "tzinfo was found to be None.",
            )
        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.DATETIME)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Datetime
        """
        return cls(literal_model.scalar.primitive.datetime)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Datetime"

    def __init__(self, value):
        """
        :param datetime.datetime value: value to wrap
        """
        super(Datetime, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(datetime=value)))

    def to_python_std(self):
        """
        :rtype: datetime.datetime
        """
        return self.scalar.primitive.datetime

    def short_string(self):
        """
        :rtype: Text
        """
        return "Datetime({})".format(_six.text_type(self.scalar.primitive.datetime))


class Timedelta(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value: Uses https://github.com/wroberts/pytimeparse for parsing
        :rtype: Timedelta
        """
        td = _parse_duration_string(string_value)
        if td is None:
            raise _user_exceptions.FlyteTypeException(
                _six.text_type,
                _datetime.timedelta,
                additional_msg="Could not convert string to" " time delta: {}".format(string_value),
            )
        return cls.from_python_std(_datetime.timedelta(seconds=td))

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif type(t_value) != _datetime.timedelta:
            raise _user_exceptions.FlyteTypeException(type(t_value), _datetime.timedelta, t_value)

        return cls(t_value)

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.DURATION)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Timedelta
        """
        return cls(literal_model.scalar.primitive.duration)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Timedelta"

    def __init__(self, value):
        """
        :param datetime.timedelta value: value to wrap
        """
        super(Timedelta, self).__init__(scalar=_literals.Scalar(primitive=_literals.Primitive(duration=value)))

    def to_python_std(self):
        """
        :rtype: datetime.timedelta
        """
        return self.scalar.primitive.duration

    def short_string(self):
        """
        :rtype: Text
        """
        return "Timedelta({})".format(self.scalar.primitive.duration)


class Generic(_base_sdk_types.FlyteSdkValue):
    @classmethod
    def from_string(cls, string_value):
        """
        :param Text string_value: Should be a JSON formatted string
        :rtype: Generic
        """
        try:
            t = _json_format.Parse(string_value, _struct.Struct())
        except Exception:
            raise _user_exceptions.FlyteValueException(string_value, "Could not be parsed from JSON.")
        return cls(t)

    @classmethod
    def is_castable_from(cls, other):
        """
        :param flytekit.common.types.base_literal_types.FlyteSdkType other:
        :rtype: bool
        """
        return cls == other

    @classmethod
    def from_python_std(cls, t_value):
        """
        :param T t_value: It is up to each individual object as to whether or not this value can be cast.
        :rtype: FlyteSdkValue
        :raises: flytekit.common.exceptions.user.FlyteTypeException
        """
        if t_value is None:
            return _base_sdk_types.Void()
        elif not isinstance(t_value, dict):
            raise _user_exceptions.FlyteTypeException(type(t_value), dict, t_value)

        try:
            t = _json.dumps(t_value)
        except Exception:
            raise _user_exceptions.FlyteValueException(t_value, "Is not JSON serializable.")

        return cls(_json_format.Parse(t, _struct.Struct()))

    @classmethod
    def to_flyte_literal_type(cls):
        """
        :rtype: flytekit.models.types.LiteralType
        """
        return _idl_types.LiteralType(simple=_idl_types.SimpleType.STRUCT)

    @classmethod
    def promote_from_model(cls, literal_model):
        """
        Creates an object of this type from the model primitive defining it.
        :param flytekit.models.literals.Literal literal_model:
        :rtype: Generic
        """
        return cls(literal_model.scalar.generic)

    @classmethod
    def short_class_string(cls):
        """
        :rtype: Text
        """
        return "Generic"

    def __init__(self, value):
        """
        :param _struct.Struct value: value to wrap
        """
        super(Generic, self).__init__(scalar=_literals.Scalar(generic=value))

    def to_python_std(self):
        """
        :rtype: dict[Text, T]
        """
        return _json.loads(_json_format.MessageToJson(self.scalar.generic))

    def short_string(self):
        """
        :rtype: Text
        """
        return "Generic({})".format(self.to_python_std())

    def long_string(self):
        """
        :rtype: Text
        """
        return "Generic({})".format(self.to_python_std())



import datetime as _datetime
import inspect
from typing import Any, Callable, Dict, List, Optional, Type, Union

from flytekit.core.base_task import TaskMetadata
from flytekit.core.interface import transform_signature_to_interface
from flytekit.core.python_function_task import PythonFunctionTask
from flytekit.core.reference_entity import ReferenceEntity, TaskReference
from flytekit.core.resources import Resources
from flytekit.models.security import Secret


class TaskPlugins(object):
    """
    This is the TaskPlugins factory for task types that are derivative of PythonFunctionTask.
    Every task that the user wishes to use should be available in this factory.
    Usage

    .. code-block:: python

        TaskPlugins.register_pythontask_plugin(config_object_type, plugin_object_type)
        # config_object_type is any class that will be passed to the plugin_object as task_config
        # Plugin_object_type is a derivative of ``PythonFunctionTask``
    """

    _PYTHONFUNCTION_TASK_PLUGINS: Dict[type, Type[PythonFunctionTask]] = {}

    @classmethod
    def register_pythontask_plugin(cls, plugin_config_type: type, plugin: Type[PythonFunctionTask]):
        """
        Use this method to register a new plugin into Flytekit. Usage ::

        .. code-block:: python

            TaskPlugins.register_pythontask_plugin(config_object_type, plugin_object_type)
            # config_object_type is any class that will be passed to the plugin_object as task_config
            # Plugin_object_type is a derivative of ``PythonFunctionTask``
        """
        if plugin_config_type in cls._PYTHONFUNCTION_TASK_PLUGINS:
            found = cls._PYTHONFUNCTION_TASK_PLUGINS[plugin_config_type]
            if found == plugin:
                return
            raise TypeError(
                f"Requesting to register plugin {plugin} - collides with existing plugin {found}"
                f" for type {plugin_config_type}"
            )

        cls._PYTHONFUNCTION_TASK_PLUGINS[plugin_config_type] = plugin

    @classmethod
    def find_pythontask_plugin(cls, plugin_config_type: type) -> Type[PythonFunctionTask]:
        """
        Returns a PluginObjectType if found or returns the base PythonFunctionTask
        """
        if plugin_config_type in cls._PYTHONFUNCTION_TASK_PLUGINS:
            return cls._PYTHONFUNCTION_TASK_PLUGINS[plugin_config_type]
        # Defaults to returning Base PythonFunctionTask
        return PythonFunctionTask


def task(
    _task_function: Optional[Callable] = None,
    task_config: Optional[Any] = None,
    cache: bool = False,
    cache_version: str = "",
    retries: int = 0,
    interruptable: bool = False,
    deprecated: str = "",
    timeout: Union[_datetime.timedelta, int] = 0,
    container_image: Optional[str] = None,
    environment: Optional[Dict[str, str]] = None,
    requests: Optional[Resources] = None,
    limits: Optional[Resources] = None,
    secret_requests: Optional[List[Secret]] = None,
    execution_mode: Optional[PythonFunctionTask.ExecutionBehavior] = PythonFunctionTask.ExecutionBehavior.DEFAULT,
) -> Union[Callable, PythonFunctionTask]:
    """
    This is the core decorator to use for any task type in flytekit.

    Tasks are the building blocks of Flyte. They represent users code. Tasks have the following properties

    * Versioned (usually tied to the git sha)
    * Strong interfaces (specified inputs and outputs)
    * Declarative
    * Independently executable
    * Unit testable

    For a simple python task,

    .. code-block:: python

        @task
        def my_task(x: int, y: typing.Dict[str, str]) -> str:
            ...

    For specific task types

    .. code-block:: python

        @task(task_config=Spark(), retries=3)
        def my_task(x: int, y: typing.Dict[str, str]) -> str:
            ...

    Please see some cookbook :std:ref:`task examples <cookbook:tasks>` for additional information.

    :param _task_function: This argument is implicitly passed and represents the decorated function
    :param task_config: This argument provides configuration for a specific task types.
                        Please refer to the plugins documentation for the right object to use.
    :param cache: Boolean that indicates if caching should be enabled
    :param cache_version: Cache version to use. Changes to the task signature will automatically trigger a cache miss,
           but you can always manually update this field as well to force a cache miss. You should also manually bump
           this version if the function body/business logic has changed, but the signature hasn't.
    :param retries: Number of times to retry this task during a workflow execution.
    :param interruptable: Boolean that indicates that this task can be interrupted and/or scheduled on nodes
                          with lower QoS guarantees. This will directly reduce the `$`/`execution cost` associated,
                          at the cost of performance penalties due to potential interruptions. Requires additional
                          Flyte platform level configuration.
    :param deprecated: A string that can be used to provide a warning message for deprecated task. Absence / empty str
                       indicates that the task is active and not deprecated
    :param timeout: the max amount of time for which one execution of this task should be executed for. The execution
                    will be terminated if the runtime exceeds the given timeout (approximately).
    :param container_image: By default the configured FLYTE_INTERNAL_IMAGE is used for every task. This directive can be
                used to provide an alternate image for a specific task. This is useful for the cases in which images
                bloat because of various dependencies and a dependency is only required for this or a set of tasks,
                and they vary from the default.

                .. code-block:: python

                    # Use default image name `fqn` and alter the tag to `tag-{{default.tag}}` tag of the default image
                    # with a prefix. In this case, it is assumed that the image like
                    # flytecookbook:tag-gitsha is published alongwith the default of flytecookbook:gitsha
                    @task(container_image='{{.images.default.fqn}}:tag-{{images.default.tag}}')
                    def foo():
                        ...

                    # Refer to configurations to configure fqns for other images besides default. In this case it will
                    # lookup for an image named xyz
                    @task(container_image='{{.images.xyz.fqn}}:{{images.default.tag}}')
                    def foo2():
                        ...
    :param environment: Environment variables that should be added for this tasks execution
    :param requests: Specify compute resource requests for your task. For Pod-plugin tasks, these values will apply only
      to the primary container.
    :param limits: Compute limits. Specify compute resource limits for your task. For Pod-plugin tasks, these values
      will apply only to the primary container. For more information, please see :py:class:`flytekit.Resources`.
    :param secret_requests: Keys that can identify the secrets supplied at runtime. Ideally the secret keys should also be
                     semi-descriptive. The key values will be available from runtime, if the backend is configured
                     to provide secrets and if secrets are available in the configured secrets store.
                     Possible options for secret stores are - Vault, Confidant, Kube secrets, AWS KMS etc
                     Refer to :py:class:`Secret` to understand how to specify the request for a secret. It
                     may change based on the backend provider.
    :param execution_mode: This is mainly for internal use. Please ignore. It is filled in automatically.
    """

    def wrapper(fn) -> PythonFunctionTask:
        _metadata = TaskMetadata(
            cache=cache,
            cache_version=cache_version,
            retries=retries,
            interruptable=interruptable,
            deprecated=deprecated,
            timeout=timeout,
        )

        task_instance = TaskPlugins.find_pythontask_plugin(type(task_config))(
            task_config,
            fn,
            metadata=_metadata,
            container_image=container_image,
            environment=environment,
            requests=requests,
            limits=limits,
            secret_requests=secret_requests,
            execution_mode=execution_mode,
        )

        return task_instance

    if _task_function:
        return wrapper(_task_function)
    else:
        return wrapper


class ReferenceTask(ReferenceEntity, PythonFunctionTask):
    """
    This is a reference task, the body of the function passed in through the constructor will never be used, only the
    signature of the function will be. The signature should also match the signature of the task you're referencing,
    as stored by Flyte Admin, if not, workflows using this will break upon compilation.
    """

    def __init__(
        self, project: str, domain: str, name: str, version: str, inputs: Dict[str, Type], outputs: Dict[str, Type]
    ):
        super().__init__(TaskReference(project, domain, name, version), inputs, outputs)

        # Reference tasks shouldn't call the parent constructor, but the parent constructor is what sets the resolver
        self._task_resolver = None


def reference_task(
    project: str,
    domain: str,
    name: str,
    version: str,
) -> Callable[[Callable[..., Any]], ReferenceTask]:
    """
    A reference task is a pointer to a task that already exists on your Flyte installation. This
    object will not initiate a network call to Admin, which is why the user is asked to provide the expected interface.
    If at registration time the interface provided causes an issue with compilation, an error will be returned.
    """

    def wrapper(fn) -> ReferenceTask:
        interface = transform_signature_to_interface(inspect.signature(fn))
        return ReferenceTask(project, domain, name, version, interface.inputs, interface.outputs)

    return wrapper



from flytekit.contrib.sensors.base_sensor import Sensor as _Sensor
from flytekit.plugins import hmsclient as _hmsclient


class _HiveSensor(_Sensor):
    def __init__(self, host, port, schema="default", **kwargs):
        """
        :param Text host:
        :param Text port:
        :param Text schema: The schema/database that we should consider.
        :param **kwargs: See flytekit.contrib.sensors.base_sensor.Sensor for more
            parameters.
        """
        self._schema = schema
        self._host = host
        self._port = port
        self._hive_metastore_client = _hmsclient.HMSClient(host=host, port=port)
        super(_HiveSensor, self).__init__(**kwargs)


class HiveTableSensor(_HiveSensor):
    def __init__(self, table_name, host, port, **kwargs):
        """
        :param Text host: The host for the Hive metastore Thrift service.
        :param Text port: The port for the Hive metastore Thrift Service
        :param Text table_name: The name of the table to look for.
        :param **kwargs: See _HiveSensor and flytekit.contrib.sensors.base_sensor.Sensor for more
            parameters.
        """
        super(HiveTableSensor, self).__init__(host, port, **kwargs)
        self._table_name = table_name

    def _do_poll(self):
        """
        :rtype: (bool, Optional[datetime.timedelta])
        """
        with self._hive_metastore_client as client:
            try:
                client.get_table(self._schema, self._table_name)
                return True, None
            except _hmsclient.genthrift.hive_metastore.ttypes.NoSuchObjectException:
                return False, None


class HiveNamedPartitionSensor(_HiveSensor):
    def __init__(self, table_name, partition_names, host, port, **kwargs):
        """
        This class allows sensing for a specific named Hive Partition.  This is the preferred partition sensing
        operator because it is more efficient than evaluating a filter expression.

        :param Text table_name: The name of the table
        :param Text partition_name: The name of the partition to listen for  (example: 'ds=2017-01-01/region=NYC')
        :param Text host: The host for the Hive metastore Thrift service.
        :param Text port: The port for the Hive metastore Thrift Service
        :param **kwargs: See _HiveSensor and flytekit.contrib.sensors.base_sensor.Sensor for more
            parameters.
        """
        super(HiveNamedPartitionSensor, self).__init__(host, port, **kwargs)
        self._table_name = table_name
        self._partition_names = partition_names

    def _do_poll(self):
        """
        :rtype: (bool, Optional[datetime.timedelta])
        """
        with self._hive_metastore_client as client:
            try:
                for partition_name in self._partition_names:
                    client.get_partition_by_name(self._schema, self._table_name, partition_name)
                return True, None
            except _hmsclient.genthrift.hive_metastore.ttypes.NoSuchObjectException:
                return False, None


class HiveFilteredPartitionSensor(_HiveSensor):
    def __init__(self, table_name, partition_filter, host, port, **kwargs):
        """
        This class allows sensing for any Hive partition that matches a filter expression.  It is recommended that the
        user should use HiveNamedPartitionSensor instead when possible because it is a more efficient API.

        :param Text table_name: The name of the table
        :param Text partition_filter: A filter expression for the partition.  (example: "ds = '2017-01-01' and
            region='NYC')
        :param Text host: The host for the Hive metastore Thrift service.
        :param Text port: The port for the Hive metastore Thrift Service
        :param **kwargs: See _HiveSensor and flytekit.contrib.sensors.base_sensor.Sensor for more
            parameters.
        """
        super(HiveFilteredPartitionSensor, self).__init__(host, port, **kwargs)
        self._table_name = table_name
        self._partition_filter = partition_filter

    def _do_poll(self):
        """
        :rtype: (bool, Optional[datetime.timedelta])
        """
        with self._hive_metastore_client as client:
            partitions = client.get_partitions_by_filter(
                db_name=self._schema,
                tbl_name=self._table_name,
                filter=self._partition_filter,
                max_parts=1,
            )
            if partitions:
                return True, None
            else:
                return False, None



import datetime
import os

from flytekitplugins.papermill import NotebookTask

from flytekit import kwtypes
from flytekit.types.file import PythonNotebook
from plugins.tests.papermilltests.testdata.datatype import X


def _get_nb_path(name: str, suffix: str = "", abs: bool = True, ext: str = ".ipynb") -> str:
    """
    Creates a correct path no matter where the test is run from
    """
    _local_path = os.path.dirname(__file__)
    path = f"{_local_path}/testdata/{name}{suffix}{ext}"
    return os.path.abspath(path) if abs else path


def test_notebook_task_simple():
    nb_name = "nb-simple"
    nb = NotebookTask(
        name="test",
        notebook_path=_get_nb_path(nb_name, abs=False),
        inputs=kwtypes(pi=float),
        outputs=kwtypes(square=float),
    )
    sqr, out, render = nb.execute(pi=4)
    assert sqr == 16.0
    assert nb.python_interface.inputs == {"pi": float}
    assert nb.python_interface.outputs.keys() == {"square", "out_nb", "out_rendered_nb"}
    assert nb.output_notebook_path == out == _get_nb_path(nb_name, suffix="-out")
    assert nb.rendered_output_path == render == _get_nb_path(nb_name, suffix="-out", ext=".html")


def test_notebook_task_multi_values():
    nb_name = "nb-multi"
    nb = NotebookTask(
        name="test",
        notebook_path=_get_nb_path(nb_name, abs=False),
        inputs=kwtypes(x=int, y=int, h=str),
        outputs=kwtypes(z=int, m=int, h=str, n=datetime.datetime),
    )
    z, m, h, n, out, render = nb.execute(x=10, y=10, h="blah")
    assert z == 20
    assert m == 100
    assert h == "blah world!"
    assert type(n) == datetime.datetime
    assert nb.python_interface.inputs == {"x": int, "y": int, "h": str}
    assert nb.python_interface.outputs.keys() == {"z", "m", "h", "n", "out_nb", "out_rendered_nb"}
    assert nb.output_notebook_path == out == _get_nb_path(nb_name, suffix="-out")
    assert nb.rendered_output_path == render == _get_nb_path(nb_name, suffix="-out", ext=".html")


def test_notebook_task_complex():
    nb_name = "nb-complex"
    nb = NotebookTask(
        name="test",
        notebook_path=_get_nb_path(nb_name, abs=False),
        inputs=kwtypes(h=str, n=int, w=str),
        outputs=kwtypes(h=str, w=PythonNotebook, x=X),
    )
    h, w, x, out, render = nb.execute(h="blah", n=10, w=_get_nb_path("nb-multi"))
    assert h == "blah world!"
    assert w is not None
    assert x.x == 10
    assert nb.python_interface.inputs == {"n": int, "h": str, "w": str}
    assert nb.python_interface.outputs.keys() == {"h", "w", "x", "out_nb", "out_rendered_nb"}
    assert nb.output_notebook_path == out == _get_nb_path(nb_name, suffix="-out")
    assert nb.rendered_output_path == render == _get_nb_path(nb_name, suffix="-out", ext=".html")



import collections as _collections
import os as _os
import uuid as _uuid

import six as _six

from flytekit.common import sdk_bases as _sdk_bases
from flytekit.common import utils as _utils
from flytekit.common.exceptions import scopes as _exception_scopes
from flytekit.common.exceptions import user as _user_exceptions
from flytekit.common.types import base_sdk_types as _base_sdk_types
from flytekit.common.types import helpers as _helpers
from flytekit.common.types import primitives as _primitives
from flytekit.common.types.impl import blobs as _blob_impl
from flytekit.configuration import sdk as _sdk_config
from flytekit.interfaces.data import data_proxy as _data_proxy
from flytekit.models import literals as _literal_models
from flytekit.models import types as _type_models
from flytekit.plugins import numpy as _np
from flytekit.plugins import pandas as _pd

# Note: For now, this is only for basic type-checking.  We need not differentiate between TINYINT, BIGINT,
# and INT or DOUBLE and FLOAT, VARCHAR and STRING, etc. as we will unpack into appropriate Python
# objects anyway.  If we work on managed tables, these more specific type specifications might become necessary.
_SUPPORTED_LITERAL_TYPE_TO_PANDAS_TYPES = None


def get_supported_literal_types_to_pandas_types():
    global _SUPPORTED_LITERAL_TYPE_TO_PANDAS_TYPES
    if _SUPPORTED_LITERAL_TYPE_TO_PANDAS_TYPES is None:
        _SUPPORTED_LITERAL_TYPE_TO_PANDAS_TYPES = {
            _primitives.Integer.to_flyte_literal_type(): {_np.int32, _np.int64, _np.uint32, _np.uint64},
            _primitives.Float.to_flyte_literal_type(): {_np.float32, _np.float64},
            _primitives.Boolean.to_flyte_literal_type(): {_np.bool},
            _primitives.Datetime.to_flyte_literal_type(): {_np.datetime64},
            _primitives.Timedelta.to_flyte_literal_type(): {_np.timedelta64},
            _primitives.String.to_flyte_literal_type(): {_np.object_, _np.str_, _np.string_},
        }
    return _SUPPORTED_LITERAL_TYPE_TO_PANDAS_TYPES


_ALLOWED_PARTITION_TYPES = {str, int}

# Hive currently has limitations where column headers are not stored when writing to an overwrite directory.  There is
# an open proposal (https://issues.apache.org/jira/browse/HIVE-12860) to improve this.  Until then, we have this
# work-around where we create an external table with the appropriate schema and write the data to our desired
# location.  The issue here is that the table information in the meta-store might not get cleaned up during a partial
# failure.
_HIVE_QUERY_FORMATTER = """
    {stage_query_str}

    CREATE TEMPORARY TABLE {table}_tmp AS {query_str};
    CREATE EXTERNAL TABLE {table} LIKE {table}_tmp STORED AS PARQUET;
    ALTER TABLE {table} SET LOCATION '{url}';

    INSERT OVERWRITE TABLE {table}
        SELECT
            {columnar_query}
        FROM {table}_tmp;
    DROP TABLE {table};
    """

# Once https://issues.apache.org/jira/browse/HIVE-12860 is resolved.  We will prefer the following syntax because it
# guarantees cleanup on partial failures.
_HIVE_QUERY_FORMATTER_V2 = """
    CREATE TEMPORARY TABLE {table} AS {query_str};

    INSERT OVERWRITE DIRECTORY '{url}' STORED AS PARQUET
        SELECT {columnar_query}
        FROM {table};
    """

# Set location in both parts of this query so in case of a partial failure, we will always have some data backing a
# partition.
_WRITE_HIVE_PARTITION_QUERY_FORMATTER = """
    ALTER TABLE {write_table} ADD IF NOT EXISTS {partition_string} LOCATION '{url}';

    ALTER TABLE {write_table} {partition_string} SET LOCATION '{url}';
    """


def _format_insert_partition_query(table_name, partition_string, remote_location):
    table_pieces = table_name.split(".")
    if len(table_pieces) > 1:
        # Hive shell commands don't allow us to alter tables and select databases in the table specification.  So
        # we split the table name and use the 'use' command to choose the correct database.
        prefix = "use {};\n".format(table_pieces[0])
        table_name = ".".join(table_pieces[1:])
    else:
        prefix = ""

    return prefix + _WRITE_HIVE_PARTITION_QUERY_FORMATTER.format(
        write_table=table_name, partition_string=partition_string, url=remote_location
    )


class _SchemaIO(object):
    def __init__(self, schema_instance, local_dir, mode):
        """
        :param Schema schema_instance:
        :param flytekit.common.utils.Directory local_dir:
        :param Text mode:
        """
        self._schema = schema_instance
        self._local_dir = local_dir
        self._chunks = []
        self._index = 0
        self._mode = mode

    def _access_guard(self):
        if not self._schema:
            raise _user_exceptions.FlyteAssertion(
                "Schema IO object has already been closed.  Cannot access chunk_count property."
            )

    @_exception_scopes.system_entry_point
    def iter_chunks(self, *args, **kwargs):
        raise _user_exceptions.FlyteAssertion("{} is write only.".format(self._schema))

    @_exception_scopes.system_entry_point
    def read(self, *args, **kwargs):
        raise _user_exceptions.FlyteAssertion("{} is write only.".format(self._schema))

    @_exception_scopes.system_entry_point
    def write(self, *args, **kwargs):
        raise _user_exceptions.FlyteAssertion("{} is read only.".format(self._schema))

    @_exception_scopes.system_entry_point
    def close(self):
        self._schema = None
        self._local_dir = None
        self._chunks = None
        self._index = 0

    @property
    @_exception_scopes.system_entry_point
    def chunk_count(self):
        self._access_guard()
        return len(self._chunks)

    @_exception_scopes.system_entry_point
    def seek(self, index):
        self._access_guard()
        if index < 0 or index > self.chunk_count:
            raise _user_exceptions.FlyteValueException(
                index,
                "Attempting to seek to a chunk that is out of range. Allowed range is [0, {}]".format(self.chunk_count),
            )
        self._index = index

    @_exception_scopes.system_entry_point
    def tell(self):
        return self._index

    def __repr__(self):
        return "{mode} IO Object for {type} @ {location}".format(
            type=self._schema.type, location=self._schema.remote_prefix, mode=self._mode
        )


class _SchemaReader(_SchemaIO):
    def __init__(self, schema_instance, local_dir):
        """
        :param Schema schema_instance:
        :param flytekit.common.utils.Directory local_dir:
        """
        super(_SchemaReader, self).__init__(schema_instance, local_dir, "Read-Only")
        self.reset_chunks()

    @_exception_scopes.system_entry_point
    def reset_chunks(self):
        self._chunks = sorted(self._local_dir.list_dir())

    @_exception_scopes.system_entry_point
    def iter_chunks(self, columns=None, **kwargs):
        self._access_guard()
        while self._index < len(self._chunks):
            chunk = self.read(columns=columns, concat=False, **kwargs)
            if chunk is not None:
                yield chunk

    @staticmethod
    def _read_parquet_with_type_promotion_override(chunk, columns, parquet_engine):
        """
        This wrapper function of pd.read_parquet() is a hack intended to fix the type promotion problem
        when using fastparquet as the underlying parquet engine.

        When using fastparquet, boolean columns containing None values will be promoted to float16 columns.
        This becomes problematic when users want to write the dataframe back into parquet
        file because float16 (halffloat) is not a supported type in parquet spec. In this function, we detect
        such columns and do override the type promotion.
        """
        df = None

        if parquet_engine == "fastparquet":
            import fastparquet.thrift_structures as _ts
            from fastparquet import ParquetFile as _ParquetFile

            # https://github.com/dask/fastparquet/issues/414#issuecomment-478983811
            df = _pd.read_parquet(chunk, columns=columns, engine=parquet_engine, index=False)
            df_column_types = df.dtypes
            pf = _ParquetFile(chunk)
            schema_column_dtypes = {l.name: l.type for l in list(pf.schema.schema_elements)}

            for idx in df_column_types[df_column_types == "float16"].index.tolist():
                # A hacky way to get the string representations of the column types of a parquet schema
                # Reference:
                # https://github.com/dask/fastparquet/blob/f4ecc67f50e7bf98b2d0099c9589c615ea4b06aa/fastparquet/schema.py
                if _ts.parquet_thrift.Type._VALUES_TO_NAMES[schema_column_dtypes[idx]] == "BOOLEAN":
                    df[idx] = df[idx].astype("object")
                    df[idx].replace({0: False, 1: True, _pd.np.nan: None}, inplace=True)

        else:
            df = _pd.read_parquet(chunk, columns=columns, engine=parquet_engine)

        return df

    @_exception_scopes.system_entry_point
    def read(self, columns=None, concat=False, truncate_extra_columns=True, **kwargs):
        """
        When this function is called, one chunk will be read and received as a Pandas data frame.  Once all chunks
        have been read, this function will return None.

        :param list[Text] columns: A list of columns to read.  They must be a subset of the columns
            defined for the Schema object.  If specified, truncate_extra_columns must be True.
        :param bool concat:  If true, the entire object will be returned in one large data frame.
        :param bool truncate_extra_columns: If true, only columns from the underlying parquet file will be read if
            they are specified as columns in the schema object (except for empty schemas which will read all columns
            regardless). If false, if there are additional columns in the underlying parquet file, they will also be
            read.
        :rtype: pandas.DataFrame
        """
        if columns is not None and truncate_extra_columns is False:
            raise _user_exceptions.FlyteAssertion(
                "When reading a schema object, it is not possible to both specify a set of columns to read and "
                "additionally not truncate_extra_columns.  Either columns must not be specified or "
                "truncate_extra_columns must be set to True (or not specified)."
            )

        self._access_guard()

        parquet_engine = _sdk_config.PARQUET_ENGINE.get()
        if parquet_engine not in {"fastparquet", "pyarrow"}:
            raise _user_exceptions.FlyteAssertion(
                "environment variable parquet_engine must be one of 'pyarrow', 'fastparquet', or be unset"
            )

        df_out = None
        if not columns:
            columns = list(self._schema.type.sdk_columns.keys())

        if len(columns) == 0 or truncate_extra_columns is False:
            columns = None

        if concat:
            frames = [
                # A hacky hack
                # TODO: follow up the issue opened in the fastparquet repo for a more general fix
                # issue URL:
                _SchemaReader._read_parquet_with_type_promotion_override(
                    chunk=chunk, columns=columns, parquet_engine=parquet_engine
                )
                # _pd.read_parquet(chunk, columns=columns, engine=parquet_engine)
                for chunk in self._chunks[self._index :]
                if _os.path.getsize(chunk) > 0
            ]
            if len(frames) == 1:
                df_out = frames[0]
            elif len(frames) > 1:
                df_out = _pd.concat(frames, copy=True)
            self._index = len(self._chunks)
        else:
            while self._index < len(self._chunks) and df_out is None:
                # Skip empty chunks so the user appears to have a continuous stream of data.
                if _os.path.getsize(self._chunks[self._index]) > 0:
                    df_out = _SchemaReader._read_parquet_with_type_promotion_override(
                        chunk=self._chunks[self._index], columns=columns, parquet_engine=parquet_engine, **kwargs
                    )
                self._index += 1

        if df_out is not None:
            self._schema.compare_dataframe_to_schema(df_out, read=True, column_subset=columns)

            # Make sure the columns are renamed to exactly what the user specifies.  This prevents unexpected
            # unicode v. string mismatches.  Also, if a schema is mapped with strict_names=False, the input might
            # have totally different names.
            user_columns = columns or _six.iterkeys(self._schema.type.sdk_columns)
            # User-specified columns may or may not be unicode
            # Since, in python 2, dictionary does a transparent translation between unicode and str for the key,
            # (https://stackoverflow.com/a/24532329)
            # we use this characteristic to create a trivial lookup dictionary, to make sure we can use either
            # unicode or str to lookup, but get back whatever type the user actually used
            user_column_dict = {c: c for c in user_columns}
            if len(self._schema.type.columns) > 0:
                # Avoid using pandas.DataFrame.rename() as this function incurs significant memory overhead
                df_out.columns = [
                    user_column_dict[col] if col in user_columns else col for col in df_out.columns.values
                ]
        return df_out


class _SchemaWriter(_SchemaIO):
    def __init__(self, schema_instance, local_dir):
        """
        :param Schema schema_instance:
        :param flytekit.common.utils.Directory local_dir:
        :param Text mode:
        """
        super(_SchemaWriter, self).__init__(schema_instance, local_dir, "Write-Only")

    @_exception_scopes.system_entry_point
    def close(self):
        """
        Closes the writing IO context and uploads data to s3.
        """
        try:
            # TODO: Introduce system logging
            # logging.info("Copying recursively {} -> {}".format(self._local_dir.name, self._schema.remote_prefix))
            _data_proxy.Data.put_data(self._local_dir.name, self._schema.remote_prefix, is_multipart=True)
        finally:
            super(_SchemaWriter, self).close()

    @_exception_scopes.system_entry_point
    def write(self, data_frame, coerce_timestamps="us", allow_truncated_timestamps=False):
        """
        Writes data frame as a chunk to the local directory owned by the Schema object.  Will later be uploaded to s3.

        :param pandas.DataFrame data_frame: data frame to write as parquet
        :param Text coerce_timestamps: format to store timestamp in parquet. 'us', 'ms', 's' are allowed values.
            Note: if your timestamps will lose data due to the coercion, your write will fail!  Nanoseconds are
            problematic in the Parquet format and will not work. See allow_truncated_timestamps.
        :param bool allow_truncated_timestamps: default False. Allow truncation when coercing timestamps to a coarser
            resolution.
        """
        self._access_guard()
        if not isinstance(data_frame, _pd.DataFrame):
            raise _user_exceptions.FlyteTypeException(
                expected_type=_pd.DataFrame,
                received_type=type(data_frame),
                received_value=data_frame,
                additional_msg="Only pandas DataFrame objects can be written to a Schema object",
            )

        self._schema.compare_dataframe_to_schema(data_frame)
        all_columns = list(data_frame.columns.values)

        # Convert all columns to unicode as pyarrow's parquet reader can not handle mixed strings and unicode.
        # Since columns from Hive are returned as unicode, if a user wants to add a column to a dataframe returned from
        # Hive, then output the new data, the user would have to provide a unicode column name which is unnatural.
        unicode_columns = [_six.text_type(col) for col in all_columns]
        data_frame.columns = unicode_columns
        try:
            filename = self._local_dir.get_named_tempfile(_os.path.join(str(self._index).zfill(6)))
            data_frame.to_parquet(
                filename,
                coerce_timestamps=coerce_timestamps,
                allow_truncated_timestamps=allow_truncated_timestamps,
            )
            if self._index == len(self._chunks):
                self._chunks.append(filename)
            self._index += 1
        finally:
            # Return to old names to prevent odd behavior with user.
            data_frame.columns = unicode_columns


class _SchemaBackingMpBlob(_blob_impl.MultiPartBlob):
    @property
    def directory(self):
        """
        :rtype: flytekit.common.utils.Directory
        """
        return self._directory

    def __enter__(self):
        if not self.local_path:
            if _data_proxy.LocalWorkingDirectoryContext.get() is None:
                raise _user_exceptions.FlyteAssertion(
                    "No temporary file system is present.  Either call this method from within the "
                    "context of a task or surround with a 'with LocalTestFileSystem():' block.  Or "
                    "specify a path when calling this function."
                )
            self._directory = _utils.AutoDeletingTempDir(
                _uuid.uuid4().hex,
                tmp_dir=_data_proxy.LocalWorkingDirectoryContext.get().name,
            )
            self._is_managed = True
            self._directory.__enter__()

            if "r" in self.mode:
                _data_proxy.Data.get_data(self.remote_location, self.local_path, is_multipart=True)

    def __exit__(self, exc_type, exc_val, exc_tb):
        if "w" in self.mode:
            _data_proxy.Data.put_data(self.local_path, self.remote_location, is_multipart=True)
        return super(_SchemaBackingMpBlob, self).__exit__(exc_type, exc_val, exc_tb)


class SchemaType(_type_models.SchemaType, metaclass=_sdk_bases.ExtendedSdkType):
    _LITERAL_TYPE_TO_PROTO_ENUM = {
        _primitives.Integer.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.INTEGER,
        _primitives.Float.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.FLOAT,
        _primitives.Boolean.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.BOOLEAN,
        _primitives.Datetime.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.DATETIME,
        _primitives.Timedelta.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.DURATION,
        _primitives.String.to_flyte_literal_type(): _type_models.SchemaType.SchemaColumn.SchemaColumnType.STRING,
    }

    def __init__(self, columns=None):
        super(SchemaType, self).__init__(None)
        self._set_columns(columns or [])

    @property
    def sdk_columns(self):
        """
        This is an ordered dictionary so iterating over it will be in the  order columns were specified in the
            constructor.
        :rtype: dict[Text, flytekit.common.types.base_sdk_types.FlyteSdkType]
        """
        return self._sdk_columns

    @property
    def columns(self):
        """
        :rtype: list[flytekit.models.types.SchemaType.SchemaColumn]
        """
        return [
            _type_models.SchemaType.SchemaColumn(n, type(self)._LITERAL_TYPE_TO_PROTO_ENUM[v.to_flyte_literal_type()])
            for n, v in _six.iteritems(self.sdk_columns)
        ]

    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.types.SchemaType model:
        :rtype: SchemaType
        """
        _PROTO_ENUM_TO_SDK_TYPE = {
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.INTEGER: _helpers.get_sdk_type_from_literal_type(
                _primitives.Integer.to_flyte_literal_type()
            ),
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.FLOAT: _helpers.get_sdk_type_from_literal_type(
                _primitives.Float.to_flyte_literal_type()
            ),
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.BOOLEAN: _helpers.get_sdk_type_from_literal_type(
                _primitives.Boolean.to_flyte_literal_type()
            ),
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.DATETIME: _helpers.get_sdk_type_from_literal_type(
                _primitives.Datetime.to_flyte_literal_type()
            ),
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.DURATION: _helpers.get_sdk_type_from_literal_type(
                _primitives.Timedelta.to_flyte_literal_type()
            ),
            _type_models.SchemaType.SchemaColumn.SchemaColumnType.STRING: _helpers.get_sdk_type_from_literal_type(
                _primitives.String.to_flyte_literal_type()
            ),
        }
        return cls([(c.name, _PROTO_ENUM_TO_SDK_TYPE[c.type]) for c in model.columns])

    def _set_columns(self, columns):
        names_seen = set()
        for column in columns:
            if not isinstance(column, tuple):
                raise _user_exceptions.FlyteValueException(
                    column,
                    "When specifying a Schema type with a known set of columns.  Each column must be "
                    "specified as a tuple in the form ('name', type).",
                )
            if len(column) != 2:
                raise _user_exceptions.FlyteValueException(
                    column,
                    "When specifying a Schema type with a known set of columns.  Each column must be "
                    "specified as a tuple in the form ('name', type).",
                )
            name, sdk_type = column
            sdk_type = _helpers.python_std_to_sdk_type(sdk_type)

            if not isinstance(name, (str, _six.text_type)):
                additional_msg = (
                    "When specifying a Schema type with a known set of columns, the first element in"
                    " each tuple must be text."
                )
                raise _user_exceptions.FlyteTypeException(
                    received_type=type(name),
                    received_value=name,
                    expected_type={str, _six.text_type},
                    additional_msg=additional_msg,
                )

            if (
                not isinstance(sdk_type, _base_sdk_types.FlyteSdkType)
                or sdk_type.to_flyte_literal_type() not in get_supported_literal_types_to_pandas_types()
            ):
                additional_msg = (
                    "When specifying a Schema type with a known set of columns, the second element of "
                    "each tuple must be a supported type.  Failed for column: {name}".format(name=name)
                )
                raise _user_exceptions.FlyteTypeException(
                    expected_type=list(get_supported_literal_types_to_pandas_types().keys()),
                    received_type=sdk_type,
                    additional_msg=additional_msg,
                )

            if name in names_seen:
                raise ValueError(
                    "The column name {name} was specified multiple times when instantiating the "
                    "Schema.".format(name=name)
                )
            names_seen.add(name)

        self._sdk_columns = _collections.OrderedDict(columns)


class Schema(_literal_models.Schema, metaclass=_sdk_bases.ExtendedSdkType):
    def __init__(self, remote_path, mode="rb", schema_type=None):
        """
        :param Text remote_path:
        :param Text mode:
        :param SchemaType schema_type: [Optional] If specified, the schema will be forced to conform to this type.  If
            not specified, the schema will be considered generic.
        """
        self._mp_blob = _SchemaBackingMpBlob(remote_path, mode=mode)
        super(Schema, self).__init__(self._mp_blob.uri, schema_type or SchemaType())
        self._io_object = None

    @classmethod
    def promote_from_model(cls, model):
        """
        :param flytekit.models.literals.Schema model:
        :rtype: Schema
        """
        return cls(model.uri, schema_type=SchemaType.promote_from_model(model.type))

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_known_location(cls, known_remote_location, mode="wb", schema_type=None):
        """
        :param Text known_remote_location: The location to which to write the object.  Usually an s3 path.
        :param Text mode:
        :param SchemaType schema_type: [Optional] If specified, the schema will be forced to conform to this type.  If
            not specified, the schema will be considered generic.
        :rtype: Schema
        """
        return cls(known_remote_location, mode=mode, schema_type=schema_type)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_at_any_location(cls, mode="wb", schema_type=None):
        """
        :param Text mode:
        :param SchemaType schema_type: [Optional] If specified, the schema will be forced to conform to this type.  If
            not specified, the schema will be considered generic.
        :rtype: Schema
        """
        return cls.create_at_known_location(_data_proxy.Data.get_remote_path(), mode=mode, schema_type=schema_type)

    @classmethod
    @_exception_scopes.system_entry_point
    def fetch(cls, remote_path, local_path=None, overwrite=False, mode="rb", schema_type=None):
        """
        :param Text remote_path: The location from which to fetch the object. Usually an s3 path.
        :param Text local_path: [Optional] A local path to which to download the object. If specified, the object
            will not be managed and might not be cleaned up by the system upon exiting the context.
        :param bool overwrite: If True, objects will be overwritten at the provided local_path in order to fetch this
            object.  Default is False.
        :param Text mode: Read or write mode of the object.
        :param SchemaType schema_type: [Optional] If specified, the schema will be forced to conform to this type.  If
            not specified, the schema will be considered generic.
        :rtype: Schema
        """
        schema = cls(remote_path, mode=mode, schema_type=schema_type)
        schema.download(local_path=local_path, overwrite=overwrite)
        return schema

    @classmethod
    @_exception_scopes.system_entry_point
    def from_python_std(cls, t_value, schema_type=None):
        """
        :param T t_value:
        :param SchemaType schema_type: [Optional] If specified, we will ensure
        :rtype: Schema
        """
        if isinstance(t_value, (str, _six.text_type)):
            if _os.path.isdir(t_value):
                schema = cls.create_at_any_location(schema_type=schema_type)
                schema.multipart_blob._directory = _utils.Directory(t_value)
                schema.upload()
            else:
                schema = cls.create_at_known_location(t_value, schema_type=schema_type)
            return schema
        elif isinstance(t_value, cls):
            return t_value
        elif isinstance(t_value, _pd.DataFrame):
            # Accepts a pandas dataframe and converts to a Schema object
            o = cls.create_at_any_location(schema_type=schema_type)
            with o as w:
                w.write(t_value)
            return o
        elif isinstance(t_value, list):
            # Accepts a list of pandas dataframe and converts to a Schema object
            o = cls.create_at_any_location(schema_type=schema_type)
            with o as w:
                for x in t_value:
                    if isinstance(x, _pd.DataFrame):
                        w.write(x)
                    else:
                        raise _user_exceptions.FlyteTypeException(
                            type(t_value),
                            {str, _six.text_type, Schema},
                            received_value=x,
                            additional_msg="A Schema object can only be create from a pandas DataFrame or a list of pandas DataFrame.",
                        )
            return o
        else:
            raise _user_exceptions.FlyteTypeException(
                type(t_value),
                {str, _six.text_type, Schema},
                received_value=t_value,
                additional_msg="Unable to create Schema from user-provided value.",
            )

    @classmethod
    def from_string(cls, string_value, schema_type=None):
        """
        :param Text string_value:
        :param SchemaType schema_type:
        :rtype: Schema
        """
        if not string_value:
            _user_exceptions.FlyteValueException(string_value, "Cannot create a Schema from an empty path")
        return cls.create_at_known_location(string_value, schema_type=schema_type)

    @classmethod
    @_exception_scopes.system_entry_point
    def create_from_hive_query(
        cls,
        select_query,
        stage_query=None,
        schema_to_table_name_map=None,
        schema_type=None,
        known_location=None,
    ):
        """
        Returns a query that can be submitted to Hive and produce the desired output.  It also returns a properly-typed
        schema object.

        :param Text select_query: Query for selecting data from Hive
        :param Text stage_query: Query for building temporary tables on Hive.
            Runs before the select query. Temporary tables are supported but CTEs are not supported.
        :param Dict[Text, Text] schema_to_table_name_map: A map of column names in the schema to the column names
            returned from the select query
        :param Text known_location: create the schema object at a known s3 location.
        :param SchemaType schema_type: [Optional] If specified, the schema will be forced to conform to this type.  If
            not specified, the schema will be considered generic.
        :return: Schema, Text
        """
        schema_object = cls(
            known_location or _data_proxy.Data.get_remote_directory(),
            mode="wb",
            schema_type=schema_type,
        )

        if len(schema_object.type.sdk_columns) > 0:
            identity_dict = {n: n for n in _six.iterkeys(schema_object.type.sdk_columns)}
            identity_dict.update(schema_to_table_name_map or {})
            schema_to_table_name_map = identity_dict

            columnar_clauses = []
            for name, sdk_type in _six.iteritems(schema_object.type.sdk_columns):
                if sdk_type == _primitives.Float:
                    columnar_clauses.append(
                        "CAST({table_column_name} as double) {schema_name}".format(
                            table_column_name=schema_to_table_name_map[name],
                            schema_name=name,
                        )
                    )
                else:
                    columnar_clauses.append(
                        "{table_column_name} as {schema_name}".format(
                            table_column_name=schema_to_table_name_map[name],
                            schema_name=name,
                        )
                    )
            columnar_query = ",\n\t\t".join(columnar_clauses)
        else:
            columnar_query = "*"

        stage_query_str = _six.text_type(stage_query or "")
        # the stage query should always end with a semicolon
        stage_query_str = stage_query_str if stage_query_str.endswith(";") else (stage_query_str + ";")
        query = _HIVE_QUERY_FORMATTER.format(
            url=schema_object.remote_location,
            stage_query_str=stage_query_str,
            query_str=select_query.strip().strip(";"),
            columnar_query=columnar_query,
            table=_uuid.uuid4().hex,
        )
        return schema_object, query

    @property
    def local_path(self):
        """
        Local filesystem path where the file was downloaded
        :rtype: Text
        """
        return self._mp_blob.local_path

    @property
    def remote_location(self):
        """
        Path to where this MultiPartBlob will be synced.  This is needed for reverse compatibility.
        :rtype: Text
        """
        return self.uri

    @property
    def remote_prefix(self):
        """
        Path to where this MultiPartBlob will be synced.  This is needed for reverse compatibility.
        :rtype: Text
        """
        return self.uri

    @property
    def uri(self):
        """
        Path to where this MultiPartBlob will be synced.
        :rtype: Text
        """
        return self.multipart_blob.uri

    @property
    def mode(self):
        """
        The mode string the MultiPartBlob is associated with.
        :rtype: Text
        """
        return self._mp_blob.mode

    @property
    def type(self):
        """
        The schema type definition associated with this object.
        :rtype: SchemaType
        """
        return self._type

    @property
    def multipart_blob(self):
        """
        :rtype: flytekit.common.types.impl.blobs.MultiPartBlob
        """
        return self._mp_blob

    @_exception_scopes.system_entry_point
    def __enter__(self):
        """
        :rtype: _SchemaIO
        """
        if self._io_object is not None:
            raise _user_exceptions.FlyteAssertion(
                "The context of a schema can only be entered once at a time.  Make sure the previous "
                "'with' block has been exited."
            )

        self._mp_blob.__enter__()
        if "r" in self.mode:
            self._io_object = _SchemaReader(self, self.multipart_blob.directory)
        else:
            self._io_object = _SchemaWriter(self, self.multipart_blob.directory)
        return self._io_object

    @_exception_scopes.system_entry_point
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._io_object = None
        return self._mp_blob.__exit__(exc_type, exc_val, exc_tb)

    def __repr__(self):
        return "Schema({columns}) @ {location} ({mode})".format(
            columns=self.type.columns,
            location=self.remote_prefix,
            mode="read-only" if "r" in self.mode else "write-only",
        )

    @_exception_scopes.system_entry_point
    def download(self, local_path=None, overwrite=False):
        """
        :param Text local_path: [Optional] A local path to which to download the object. If specified, the object
            will not be managed and might not be cleaned up by the system upon exiting the context.
        :param bool overwrite: If True, objects will be overwritten at the provided local_path in order to fetch this
            object.  Default is False.
        :rtype: Schema
        """
        self.multipart_blob.download(local_path=local_path, overwrite=overwrite)

    @_exception_scopes.system_entry_point
    def get_write_partition_to_hive_table_query(
        self,
        table_name,
        partitions=None,
        schema_to_table_name_map=None,
        partitions_in_table=False,
        append_to_partition=False,
    ):
        """
        Returns a Hive query string that will update the metatable to point to the data as the new partition.

        :param Text table_name:
        :param dict[Text, T] partitions: A dictionary mapping table partition key names to the values matching this
            partition.
        :param dict[Text, Text] schema_to_table_name_map: Mapping of names in current schema to table in which it is
            being inserted.  Currently not supported.  Must be None.
        :param bool partitions_in_table:  Whether or not the partition columns exist in the data being submitted.
            Currently not supported.  Must be false
        :param bool append_to_partition: Whether or not to append new values to a partition.  Currently not supported.
        :return: Text
        """
        partition_string = ""
        where_string = ""
        identity_dict = {n: n for n in _six.iterkeys(self.type.sdk_columns)}
        identity_dict.update(schema_to_table_name_map or {})
        schema_to_table_name_map = identity_dict
        table_to_schema_name_map = {v: k for k, v in _six.iteritems(schema_to_table_name_map)}

        if partitions:
            partition_conditions = []
            for partition_name, partition_value in _six.iteritems(partitions):
                if not isinstance(partition_name, (str, _six.text_type)):
                    raise _user_exceptions.FlyteTypeException(
                        expected_type={str, _six.text_type},
                        received_type=type(partition_name),
                        received_value=partition_name,
                        additional_msg="All partition names must be type str.",
                    )
                if type(partition_value) not in _ALLOWED_PARTITION_TYPES:
                    raise _user_exceptions.FlyteTypeException(
                        expected_type=_ALLOWED_PARTITION_TYPES,
                        received_type=type(partition_value),
                        received_value=partition_value,
                        additional_msg="Partition {name} has an unsupported type.".format(name=partition_name),
                    )

                # We need the string to be quoted in the query, so let's take repr of it.
                if isinstance(partition_value, (str, _six.text_type)):
                    partition_value = repr(partition_value)
                partition_conditions.append(
                    "{partition_name} = {partition_value}".format(
                        partition_name=partition_name, partition_value=partition_value
                    )
                )
            partition_formatter = "PARTITION (\n\t{conditions}\n)"
            partition_string = partition_formatter.format(conditions=",\n\t".join(partition_conditions))

        if partitions_in_table and partitions:
            where_clauses = []
            for partition_name, partition_value in partitions:
                where_clauses.append(
                    "\n\t\t{schema_name} = {value_str} AND ".format(
                        schema_name=table_to_schema_name_map[partition_name],
                        value_str=partition_value,
                    )
                )
            where_string = "WHERE\n\t\t{where_clauses}".format(where_clauses=" AND\n\t\t".join(where_clauses))

        if where_string or partitions_in_table:
            raise _user_exceptions.FlyteAssertion(
                "Currently, the partition values should not be present in the schema pushed to Hive."
            )
        if append_to_partition:
            raise _user_exceptions.FlyteAssertion(
                "Currently, partitions can only be overwritten, they cannot be appended."
            )
        if not partitions:
            raise _user_exceptions.FlyteAssertion(
                "Currently, partition values MUST be specified for writing to a table."
            )

        return _format_insert_partition_query(
            remote_location=self.remote_location,
            table_name=table_name,
            partition_string=partition_string,
        )

    def compare_dataframe_to_schema(self, data_frame, column_subset=None, read=False):
        """
        Do necessary type checking of a pandas data frame.  Raise exception if it doesn't match.
        :param pandas.DateFrame data_frame: data frame to type check
        :param list[Text] column_subset:
        :param bool read: Used to alter error message for more clarity.
        """
        all_columns = list(data_frame.columns.values)
        schema_column_names = list(self.type.sdk_columns.keys())

        # Skip checking if we have a generic schema type (no specified columns)
        if not schema_column_names:
            return

        # If we specify a subset of columns, ensure they all exist and then only take those columns
        if column_subset is not None:
            schema_column_names = []
            failed_columns = []
            for column in column_subset:
                if column not in self.type.sdk_columns:
                    failed_columns.append(column)
                else:
                    schema_column_names.append(column)

            if len(failed_columns) > 0:
                additional_msg = ""
                raise _user_exceptions.FlyteAssertion(
                    "{} was/where requested but could not be found in the schema: {}.{}".format(
                        failed_columns, self.type.sdk_columns, additional_msg
                    )
                )

        if not all(c in all_columns for c in schema_column_names):
            raise _user_exceptions.FlyteTypeException(
                expected_type=self.type.sdk_columns,
                received_type=data_frame.columns,
                additional_msg="Mismatch between the data frame's column names {} and schema's column names {} "
                "with strict_names=True.".format(all_columns, schema_column_names),
            )

        # This only iterates if the Schema has specified columns.
        for name in schema_column_names:
            literal_type = self.type.sdk_columns[name].to_flyte_literal_type()
            dtype = data_frame[name].dtype

            # TODO np.issubdtype is deprecated. Replace it
            if all(
                not _np.issubdtype(dtype, allowed_type)
                for allowed_type in get_supported_literal_types_to_pandas_types()[literal_type]
            ):
                if read:
                    read_or_write_msg = "read data frame object from schema"
                else:
                    read_or_write_msg = "write data frame object to schema"
                additional_msg = (
                    "Cannot {read_write} because the types do not match. Column "
                    "'{name}' did not pass type checking.  Note: If your "
                    "column contains null values, the types might not transition as expected between parquet and "
                    "pandas.  For more information, see: "
                    "http://arrow.apache.org/docs/python/pandas.html#arrow-pandas-conversion".format(
                        read_write=read_or_write_msg, name=name
                    )
                )
                raise _user_exceptions.FlyteTypeException(
                    expected_type=get_supported_literal_types_to_pandas_types()[literal_type],
                    received_type=dtype,
                    additional_msg=additional_msg,
                )

    def cast_to(self, other_type):
        """
        :param SchemaType other_type:
        :rtype: Schema
        """
        if len(other_type.sdk_columns) > 0:
            for k, v in _six.iteritems(other_type.sdk_columns):
                if k not in self.type.sdk_columns:
                    raise _user_exceptions.FlyteTypeException(
                        self.type,
                        other_type,
                        additional_msg="Cannot cast because a required column '{}' was not found.".format(k),
                        received_value=self,
                    )
                if (
                    not isinstance(v, _base_sdk_types.FlyteSdkType)
                    or v.to_flyte_literal_type() != self.type.sdk_columns[k].to_flyte_literal_type()
                ):
                    raise _user_exceptions.FlyteTypeException(
                        self.type.sdk_columns[k],
                        v,
                        additional_msg="Cannot cast because the column type for column '{}' does not match.".format(k),
                    )
        return Schema(self.remote_location, mode=self.mode, schema_type=other_type)

    @_exception_scopes.system_entry_point
    def upload(self):
        """
        Upload the schema to the remote location
        """
        if "w" not in self.mode:
            raise _user_exceptions.FlyteAssertion("Cannot upload a read-only schema!")

        elif not self.local_path:
            raise _user_exceptions.FlyteAssertion(
                "The schema is not currently backed by a local directory "
                "and therefore cannot be uploaded.  Please write to this before "
                "attempting an upload."
            )
        else:
            # TODO: Introduce system logging
            # logging.info("Putting {} -> {}".format(self.local_path, self.remote_location))
            _data_proxy.Data.put_data(self.local_path, self.remote_location, is_multipart=True)



import re

from flytekit.configuration import common as _common_config

IMAGE = _common_config.FlyteRequiredStringConfigurationEntry("internal", "image")
# This configuration option specifies the path to the file that holds the configuration options.  Don't worry,
# there will not be cycles because the parsing of the configuration file intentionally will not read and settings
# in the [internal] section.
# The default, if you want to use it, should be a file called flytekit.config, located in wherever your python
# interpreter originates.
CONFIGURATION_PATH = _common_config.FlyteStringConfigurationEntry(
    "internal", "configuration_path", default="flytekit.config"
)

# Project, Domain and Version represent the values at registration time.
PROJECT = _common_config.FlyteStringConfigurationEntry("internal", "project", default="")
DOMAIN = _common_config.FlyteStringConfigurationEntry("internal", "domain", default="")
NAME = _common_config.FlyteStringConfigurationEntry("internal", "name", default="")
VERSION = _common_config.FlyteStringConfigurationEntry("internal", "version", default="")

# Project, Domain and Version represent the values at registration time.
TASK_PROJECT = _common_config.FlyteStringConfigurationEntry("internal", "task_project", default="")
TASK_DOMAIN = _common_config.FlyteStringConfigurationEntry("internal", "task_domain", default="")
TASK_NAME = _common_config.FlyteStringConfigurationEntry("internal", "task_name", default="")
TASK_VERSION = _common_config.FlyteStringConfigurationEntry("internal", "task_version", default="")

# Execution project and domain represent the values passed by execution engine at runtime.
EXECUTION_PROJECT = _common_config.FlyteStringConfigurationEntry("internal", "execution_project", default="")
EXECUTION_DOMAIN = _common_config.FlyteStringConfigurationEntry("internal", "execution_domain", default="")
EXECUTION_WORKFLOW = _common_config.FlyteStringConfigurationEntry("internal", "execution_workflow", default="")
EXECUTION_LAUNCHPLAN = _common_config.FlyteStringConfigurationEntry("internal", "execution_launchplan", default="")
EXECUTION_NAME = _common_config.FlyteStringConfigurationEntry("internal", "execution_id", default="")

# This is another layer of logging level, which can be set by propeller, and can override the SDK configuration if
# necessary.  (See the sdk.py version of this as well.)
LOGGING_LEVEL = _common_config.FlyteIntegerConfigurationEntry("internal", "logging_level")

_IMAGE_VERSION_REGEX = ".*:(.+)"


def look_up_version_from_image_tag(tag):
    """
    Looks up the image tag from environment variable (should be set from the Dockerfile).
        FLYTE_INTERNAL_IMAGE should be the environment variable.

    This function is used when registering tasks/workflows with Admin.
    When using the canonical Python-based development cycle, the version that is used to register workflows
    and tasks with Admin should be the version of the image itself, which should ideally be something unique
    like the sha of the latest commit.

    :param Text tag: e.g. somedocker.com/myimage:someversion123
    :rtype: Text
    """
    if tag is None or tag == "":
        raise Exception("Bad input for image tag {}".format(tag))
    m = re.match(_IMAGE_VERSION_REGEX, tag)
    if m is not None:
        return m.group(1)

    raise Exception("Could not parse image version from configuration. Did you set it in the" "Dockerfile?")



import abc as _abc


class DataProxy(object, metaclass=_abc.ABCMeta):
    def exists(self, path):
        """
        :param path:
        :rtype: bool: whether the file exists or not
        """
        pass

    def download_directory(self, remote_path, local_path):
        """
        :param Text remote_path:
        :param Text local_path:
        """
        pass

    def download(self, remote_path, local_path):
        """
        :param Text remote_path:
        :param Text local_path:
        """
        pass

    def upload(self, file_path, to_path):
        """
        :param Text file_path:
        :param Text to_path:
        """
        pass

    def upload_directory(self, local_path, remote_path):
        """
        :param Text local_path:
        :param Text remote_path:
        """
        pass

    def get_random_path(self):
        """
        :rtype: Text
        """
        pass

    def get_random_directory(self):
        """
        :rtype: Text
        """
        pass



import abc as _abc
import configparser as _configparser
import os as _os

from flytekit.common.exceptions import user as _user_exceptions


def format_section_key(section, key):
    """
    :param Text section:
    :param Text key:
    :rtype: Text
    """
    return "FLYTE_{section}_{key}".format(section=section.upper(), key=key.upper())


class FlyteConfigurationFile(object):
    def __init__(self, location=None):
        """
        This singleton is initialized on module load with empty location. If pyflyte is called with
        a config flag, it'll reload the singleton with the passed config path.

        :param Text location: used to load config from this location.
        """
        self._location = None
        self._config = None
        self.reset_config(location)

    def _load_config(self):
        if self._config is None and self._location:
            config = _configparser.ConfigParser()
            config.read(self._location)
            if config.has_section("internal"):
                raise _user_exceptions.FlyteAssertion(
                    "The config file '{}' cannot contain a section for internal "
                    "only configurations.".format(self._location)
                )
            self._config = config

    def get_string(self, section, key, default=None):
        """
        :param Text section:
        :param Text key:
        :param Text default:
        :rtype: Text
        """
        self._load_config()
        if self._config is not None:
            try:
                return self._config.get(section, key, fallback=default)
            except Exception:
                pass
        return default

    def get_int(self, section, key, default=None):
        """
        :param Text section:
        :param Text key:
        :param int default:
        :rtype: int
        """
        self._load_config()
        if self._config is not None:
            try:
                return self._config.getint(section, key, fallback=default)
            except Exception:
                pass
        return default

    def get_bool(self, section, key, default=None):
        """
        :param Text section:
        :param Text key:
        :param bool default:
        :rtype: bool
        """
        self._load_config()
        if self._config is not None:
            try:
                return self._config.getboolean(section, key, fallback=default)
            except Exception:
                pass
        return default

    def reset_config(self, location):
        """
        :param Text location:
        """
        self._location = location or _os.environ.get("FLYTE_INTERNAL_CONFIGURATION_PATH")
        self._config = None

    @property
    def config(self) -> _configparser.ConfigParser:
        self._load_config()
        return self._config


class _FlyteConfigurationPatcher(object):
    def __init__(self, new_value, config):
        """
        :param Text new_value:
        :param _FlyteConfigurationEntry config:
        """
        self._new_value = new_value
        self._config = config
        self._old_value = None

    def __enter__(self):
        self._old_value = _os.environ.get(self._config.env_var, None)
        if self._new_value is not None:
            _os.environ[self._config.env_var] = self._new_value
        elif self._old_value is not None:
            del _os.environ[self._config.env_var]

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._old_value is not None:
            _os.environ[self._config.env_var] = self._old_value
        else:
            del _os.environ[self._config.env_var]


def _get_file_contents(location):
    """
    This reads an input file, and returns the string contents, and should be used for reading credentials.
    This function will also strip newlines.

    :param Text location: The file path holding the client id or secret
    :rtype: Text
    """
    if _os.path.isfile(location):
        with open(location, "r") as f:
            return f.read().replace("\n", "")
    return None


class _FlyteConfigurationEntry(object, metaclass=_abc.ABCMeta):
    def __init__(self, section, key, default=None, validator=None, fallback=None):
        self._section = section
        self._key = key
        self._default = default
        self._validator = validator
        self._fallback = fallback

    @property
    def env_var(self):
        """
        :rtype: Text
        """
        return format_section_key(self._section, self._key)

    @_abc.abstractmethod
    def _getter(self):
        pass

    def retrieve_value(self):
        """
        The logic in this function changes the lookup behavior for all configuration objects before hitting the
        configuration file.

        For a given configuration object ('mysection', 'mysetting'), it will now look at this waterfall:

            i.) The environment variable named 'FLYTE_MYSECTION_MYSETTING'

            ii.) The value of the environment variable that is named the value of the environment variable named
                 'FLYTE_MYSECTION_MYSETTING'. That is if os.environ['FLYTE_MYSECTION_MYSETTING'] = 'AAA' and
                  os.environ['AA'] = 'abc', then 'abc' will be the final value.

            iii.) The contents of the file pointed to by the environment variable named 'FLYTE_MYSECTION_MYSETTING',
                  assuming the value is a file.

        While it is helpful, this pattern does interrupt the manually specified fallback logic, by effective injecting
        two more fallbacks behind the scenes. Just keep this in mind as you are using/creating configuration objects.
        :rtype: Text
        """
        val = _os.environ.get(self.env_var, None)
        if val is None:
            referenced_env_var = _os.environ.get("{}_FROM_ENV_VAR".format(self.env_var), None)
            if referenced_env_var is not None:
                val = _os.environ.get(referenced_env_var, None)
            if val is None:
                referenced_file = _os.environ.get("{}_FROM_FILE".format(self.env_var), None)
                if referenced_file is not None:
                    val = _get_file_contents(referenced_file)
        return val

    def get(self):
        val = self._getter()
        if val is None and self._fallback is not None:
            val = self._fallback.get()
        if self._validator:
            self._validator(val)
        return val

    def is_set(self):
        val = self._getter()
        return val is not None

    def get_patcher(self, value):
        return _FlyteConfigurationPatcher(value, self)


class _FlyteRequiredConfigurationEntry(_FlyteConfigurationEntry):
    def __init__(self, section, key, validator=None):
        super(_FlyteRequiredConfigurationEntry, self).__init__(section, key, validator=self._validate_not_null)
        self._extra_validator = validator

    def _validate_not_null(self, val):
        if val is None:
            raise _user_exceptions.FlyteAssertion(
                "No configuration set for [{}] {}.  This is a required configuration.".format(self._section, self._key)
            )
        if self._extra_validator:
            self._extra_validator(val)


class FlyteStringConfigurationEntry(_FlyteConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_string(self._section, self._key, default=self._default)
        return val


class FlyteIntegerConfigurationEntry(_FlyteConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_int(self._section, self._key, default=self._default)
        if val is not None:
            return int(val)
        return None


class FlyteBoolConfigurationEntry(_FlyteConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()

        if val is None:
            return CONFIGURATION_SINGLETON.get_bool(self._section, self._key, default=self._default)
        else:
            # Because bool('False') is True, compare to the same values that ConfigParser uses
            if val.lower() in ["false", "0", "off", "no"]:
                return False
            return True


class FlyteStringListConfigurationEntry(_FlyteConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_string(self._section, self._key)
        if val is None:
            return self._default
        return val.split(",")


class FlyteRequiredStringConfigurationEntry(_FlyteRequiredConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_string(self._section, self._key, default=self._default)
        return val


class FlyteRequiredIntegerConfigurationEntry(_FlyteRequiredConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_int(self._section, self._key, default=self._default)
        return int(val)


class FlyteRequiredBoolConfigurationEntry(_FlyteRequiredConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_bool(self._section, self._key, default=self._default)
        return bool(val)


class FlyteRequiredStringListConfigurationEntry(_FlyteRequiredConfigurationEntry):
    def _getter(self):
        val = self.retrieve_value()
        if val is None:
            val = CONFIGURATION_SINGLETON.get_string(self._section, self._key)
        if val is None:
            return self._default
        return val.split(",")


CONFIGURATION_SINGLETON = FlyteConfigurationFile()
